from fal.cli.apps import (
    _delete,
    _delete_rev,
    _list,
    _list_rev,
    _runners,
    _scale,
    _set_rev,
)
from fal.cli.main import parse_args


def test_list():
    args = parse_args(["apps", "list"])
    assert args.func == _list


def test_list_rev():
    args = parse_args(["apps", "list-rev"])
    assert args.func == _list_rev


def test_set_rev():
    args = parse_args(["apps", "set-rev", "myapp", "myrev", "--auth", "public"])
    assert args.func == _set_rev
    assert args.app_name == "myapp"
    assert args.app_rev == "myrev"
    assert args.auth == "public"


def test_scale():
    args = parse_args(
        [
            "apps",
            "scale",
            "myapp",
            "--keep-alive",
            "123",
            "--max-multiplexing",
            "321",
            "--min-concurrency",
            "7",
            "--max-concurrency",
            "10",
            "--concurrency-buffer",
            "3",
            "--concurrency-buffer-perc",
            "27",
        ]
    )
    assert args.func == _scale
    assert args.app_name == "myapp"
    assert args.keep_alive == 123
    assert args.max_multiplexing == 321
    assert args.min_concurrency == 7
    assert args.max_concurrency == 10
    assert args.concurrency_buffer == 3
    assert args.concurrency_buffer_perc == 27


def test_runners():
    args = parse_args(["apps", "runners", "myapp"])
    assert args.func == _runners


def test_delete():
    args = parse_args(["apps", "delete", "myapp"])
    assert args.func == _delete
    assert args.app_name == "myapp"


def test_delete_rev():
    args = parse_args(["apps", "delete-rev", "myrev"])
    assert args.func == _delete_rev
    assert args.app_rev == "myrev"



from __future__ import annotations

from dataclasses import dataclass

import fal._serialization
from fal import App, wrap_app

from .api import FalServerlessError, FalServerlessHost, IsolatedFunction


@dataclass
class LoadedFunction:
    function: IsolatedFunction
    endpoints: list[str]
    app_name: str | None
    app_auth: str | None
    source_code: str | None


def load_function_from(
    host: FalServerlessHost,
    file_path: str,
    function_name: str | None = None,
) -> LoadedFunction:
    import os
    import runpy
    import sys

    sys.path.append(os.getcwd())
    module = runpy.run_path(file_path)
    if function_name is None:
        fal_objects = {
            obj_name: obj
            for obj_name, obj in module.items()
            if isinstance(obj, type) and issubclass(obj, fal.App) and obj is not fal.App
        }
        if len(fal_objects) == 0:
            raise FalServerlessError("No fal.App found in the module.")
        elif len(fal_objects) > 1:
            raise FalServerlessError(
                "Multiple fal.Apps found in the module. "
                "Please specify the name of the app."
            )

        [(function_name, obj)] = fal_objects.items()
        app_name = obj.app_name
        app_auth = obj.app_auth
    else:
        app_name = None
        app_auth = None

    if function_name not in module:
        raise FalServerlessError(f"Function '{function_name}' not found in module")

    # The module for the function is set to <run_path> when runpy is used, in which
    # case we want to manually include the package it is defined in.
    fal._serialization.include_package_from_path(file_path)

    target = module[function_name]

    with open(file_path) as f:
        source_code = f.read()

    endpoints = ["/"]
    if isinstance(target, type) and issubclass(target, App):
        app_name = target.app_name
        app_auth = target.app_auth
        endpoints = target.get_endpoints() or ["/"]
        target = wrap_app(target, host=host)

    if not isinstance(target, IsolatedFunction):
        raise FalServerlessError(
            f"Function '{function_name}' is not a fal.function or a fal.App"
        )
    return LoadedFunction(
        target, endpoints, app_name=app_name, app_auth=app_auth, source_code=source_code
    )



from __future__ import annotations

import enum
from contextlib import ExitStack
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Generic,
    Iterator,
    Literal,
    Optional,
    TypeVar,
)

import grpc
import isolate_proto
from isolate.server.interface import from_grpc, to_serialized_object, to_struct

from fal import flags
from fal._serialization import patch_pickle
from fal.auth import UserAccess, key_credentials
from fal.logging import get_logger
from fal.logging.trace import TraceContextInterceptor

if TYPE_CHECKING:
    from isolate.logs import Log

ResultT = TypeVar("ResultT")
InputT = TypeVar("InputT")
UNSET = object()

_DEFAULT_SERIALIZATION_METHOD = "cloudpickle"
FAL_SERVERLESS_DEFAULT_KEEP_ALIVE = 10
FAL_SERVERLESS_DEFAULT_MAX_MULTIPLEXING = 1
FAL_SERVERLESS_DEFAULT_MIN_CONCURRENCY = 0
FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER = 0
FAL_SERVERLESS_DEFAULT_CONCURRENCY_BUFFER_PERC = 0
ALIAS_AUTH_MODES = ["public", "private", "shared"]

logger = get_logger(__name__)

patch_pickle()


AuthModeLiteral = Literal["public", "private", "shared"]
DeploymentStrategyLiteral = Literal["recreate", "rolling"]


class ServerCredentials:
    def to_grpc(self) -> grpc.ChannelCredentials:
        raise NotImplementedError

    @property
    def base_options(self) -> dict[str, str | int]:
        import json

        from isolate_proto.configuration import GRPC_OPTIONS

        grpc_ops: dict[str, str | int] = dict(GRPC_OPTIONS)
        grpc_ops["grpc.enable_retries"] = 1
        grpc_ops["grpc.service_config"] = json.dumps(
            {
                "methodConfig": [
                    {
                        "name": [{}],
                        "retryPolicy": {
                            "maxAttempts": 5,
                            "initialBackoff": "0.1s",
                            "maxBackoff": "5s",
                            "backoffMultiplier": 2,
                            "retryableStatusCodes": ["UNAVAILABLE"],
                        },
                    }
                ]
            }
        )

        return grpc_ops


class LocalCredentials(ServerCredentials):
    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.local_channel_credentials()


class RemoteCredentials(ServerCredentials):
    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.ssl_channel_credentials()


@dataclass
class _GRPCMetadata(grpc.AuthMetadataPlugin):
    """Key value metadata bundle for gRPC credentials"""

    _key: str
    _value: str

    def __call__(
        self,
        context: grpc.AuthMetadataContext,
        callback: grpc.AuthMetadataPluginCallback,
    ) -> None:
        callback(((self._key, self._value),), None)


def get_default_server_credentials() -> ServerCredentials:
    if flags.TEST_MODE:
        return LocalCredentials()
    else:
        return RemoteCredentials()


class Credentials:
    # Cannot use `field` because child classes don't have default for all properties.
    server_credentials: ServerCredentials = get_default_server_credentials()

    def to_grpc(self) -> grpc.ChannelCredentials:
        return self.server_credentials.to_grpc()

    def to_headers(self) -> dict[str, str]:
        return {}


@dataclass
class FalServerlessKeyCredentials(Credentials):
    key_id: str
    key_secret: str

    def to_grpc(self) -> grpc.ChannelCredentials:
        return grpc.composite_channel_credentials(
            self.server_credentials.to_grpc(),
            grpc.metadata_call_credentials(_GRPCMetadata("auth-key", self.key_secret)),
            grpc.metadata_call_credentials(_GRPCMetadata("auth-key-id", self.key_id)),
        )

    def to_headers(self) -> dict[str, str]:
        return {"Authorization": f"Key {self.key_id}:{self.key_secret}"}


@dataclass
class AuthenticatedCredentials(Credentials):
    user: UserAccess = field(default_factory=UserAccess)
    team: str | None = None

    def to_grpc(self) -> grpc.ChannelCredentials:
        creds = [
            self.server_credentials.to_grpc(),
            grpc.access_token_call_credentials(self.user.access_token),
        ]

        if self.team:
            team_id = self.user.get_account(self.team)["user_id"]
            creds.append(
                grpc.metadata_call_credentials(_GRPCMetadata("fal-user-id", team_id))
            )

        return grpc.composite_channel_credentials(*creds)

    def to_headers(self) -> dict[str, str]:
        token = self.user.bearer_token
        headers = {
            "Authorization": token,
        }

        if self.team:
            team_id = self.user.get_account(self.team)["user_id"]
            headers["X-Fal-User-Id"] = team_id

        return headers


@dataclass
class ServerlessSecret:
    name: str
    created_at: datetime


def get_agent_credentials(original_credentials: Credentials) -> Credentials:
    """If running inside a fal Serverless box, use the preconfigured credentials
    instead of the user provided ones."""

    from isolate.connections.common import is_agent

    key_creds = key_credentials()
    if is_agent() and key_creds:
        return FalServerlessKeyCredentials(key_creds[0], key_creds[1])
    else:
        return original_credentials


def get_default_credentials(team: str | None = None) -> Credentials:
    from fal.config import Config

    if flags.AUTH_DISABLED:
        return Credentials()

    key_creds = key_credentials()
    if key_creds:
        logger.debug("Using key credentials")
        if team:
            raise ValueError("Using explicit team with key credentials is not allowed")
        return FalServerlessKeyCredentials(key_creds[0], key_creds[1])
    else:
        config = Config()
        team = team or config.get_internal("team")
        return AuthenticatedCredentials(team=team)


@dataclass
class FalServerlessClient:
    hostname: str
    credentials: Credentials = field(default_factory=get_default_credentials)

    def connect(self) -> FalServerlessConnection:
        return FalServerlessConnection(self.hostname, self.credentials)


class HostedRunState(Enum):
    IN_PROGRESS = 0
    SUCCESS = 1
    INTERNAL_FAILURE = 2


@dataclass
class HostedRunStatus:
    state: HostedRunState


@dataclass
class File:
    hash: str
    relative_path: str


@dataclass
class ApplicationInfo:
    application_id: str
    keep_alive: int
    max_concurrency: int
    max_multiplexing: int
    active_runners: int
    min_concurrency: int
    concurrency_buffer: int
    concurrency_buffer_perc: int
    machine_types: list[str]
    request_timeout: int
    startup_timeout: int
    valid_regions: list[str]
    created_at: datetime


@dataclass
class AliasInfo:
    alias: str
    revision: str
    auth_mode: str
    keep_alive: int
    max_concurrency: int
    max_multiplexing: int
    active_runners: int
    min_concurrency: int
    concurrency_buffer: int
    concurrency_buffer_perc: int
    machine_types: list[str]
    request_timeout: int
    startup_timeout: int
    valid_regions: list[str]


class RunnerState(Enum):
    RUNNING = "running"
    PENDING = "pending"
    SETUP = "setup"
    DOCKER_PULL = "docker_pull"
    DEAD = "dead"
    UNKNOWN = "unknown"

    @staticmethod
    def from_proto(proto: isolate_proto.RunnerInfo.State) -> RunnerState:
        if proto is isolate_proto.RunnerInfo.State.RUNNING:
            return RunnerState.RUNNING
        elif proto is isolate_proto.RunnerInfo.State.PENDING:
            return RunnerState.PENDING
        elif proto is isolate_proto.RunnerInfo.State.SETUP:
            return RunnerState.SETUP
        elif proto is isolate_proto.RunnerInfo.State.DEAD:
            return RunnerState.DEAD
        elif proto is isolate_proto.RunnerInfo.State.DOCKER_PULL:
            return RunnerState.DOCKER_PULL
        else:
            return RunnerState.UNKNOWN


@dataclass
class RunnerInfo:
    runner_id: str
    in_flight_requests: int
    expiration_countdown: Optional[int]
    uptime: timedelta
    external_metadata: dict[str, Any]
    revision: str
    alias: str
    state: RunnerState


@dataclass
class ServiceURLs:
    playground: str
    run: str
    queue: str
    ws: str


@dataclass
class HostedRunResult(Generic[ResultT]):
    run_id: str
    status: HostedRunStatus
    logs: list[Log] = field(default_factory=list)
    result: ResultT | None = None
    stream: Any = None
    service_urls: ServiceURLs | None = None


@dataclass
class RegisterApplicationResult:
    result: RegisterApplicationResultType | None
    logs: list[Log] = field(default_factory=list)
    service_urls: ServiceURLs | None = None


@dataclass
class RegisterApplicationResultType:
    application_id: str


@dataclass
class UserKeyInfo:
    key_id: str
    created_at: datetime
    scope: KeyScope
    alias: str


@dataclass
class WorkerStatus:
    worker_id: str
    start_time: datetime
    end_time: datetime
    duration: timedelta
    user_id: str
    machine_type: str


class KeyScope(enum.Enum):
    ADMIN = "ADMIN"
    API = "API"

    @staticmethod
    def from_proto(
        proto: isolate_proto.CreateUserKeyRequest.Scope.ValueType | None,
    ) -> KeyScope:
        if proto is None:
            return KeyScope.API

        if proto is isolate_proto.CreateUserKeyRequest.Scope.ADMIN:
            return KeyScope.ADMIN
        elif proto is isolate_proto.CreateUserKeyRequest.Scope.API:
            return KeyScope.API
        else:
            raise ValueError(f"Unknown KeyScope: {proto}")


class DeploymentStrategy(enum.Enum):
    RECREATE = "recreate"
    ROLLING = "rolling"

    @staticmethod
    def from_proto(
        proto: isolate_proto.DeploymentStrategy.ValueType | None,
    ) -> DeploymentStrategy:
        if proto is None:
            return DeploymentStrategy.RECREATE

        if proto is isolate_proto.DeploymentStrategy.RECREATE:
            return DeploymentStrategy.RECREATE
        elif proto is isolate_proto.DeploymentStrategy.ROLLING:
            return DeploymentStrategy.ROLLING
        else:
            raise ValueError(f"Unknown DeploymentStrategy: {proto}")

    def to_proto(self) -> isolate_proto.DeploymentStrategy.ValueType:
        if self is DeploymentStrategy.RECREATE:
            return isolate_proto.DeploymentStrategy.RECREATE
        elif self is DeploymentStrategy.ROLLING:
            return isolate_proto.DeploymentStrategy.ROLLING
        else:
            raise ValueError(f"Unknown DeploymentStrategy: {self}")


@from_grpc.register(isolate_proto.ApplicationInfo)
def _from_grpc_application_info(
    message: isolate_proto.ApplicationInfo,
) -> ApplicationInfo:
    return ApplicationInfo(
        application_id=message.application_id,
        keep_alive=message.keep_alive,
        max_concurrency=message.max_concurrency,
        max_multiplexing=message.max_multiplexing,
        active_runners=message.active_runners,
        min_concurrency=message.min_concurrency,
        concurrency_buffer=message.concurrency_buffer,
        concurrency_buffer_perc=message.concurrency_buffer_perc,
        machine_types=list(message.machine_types),
        request_timeout=message.request_timeout,
        startup_timeout=message.startup_timeout,
        valid_regions=list(message.valid_regions),
        created_at=isolate_proto.datetime_from_timestamp(message.created_at),
    )


@from_grpc.register(isolate_proto.AliasInfo)
def _from_grpc_alias_info(message: isolate_proto.AliasInfo) -> AliasInfo:
    if message.auth_mode is isolate_proto.ApplicationAuthMode.PUBLIC:
        auth_mode = "public"
    elif message.auth_mode is isolate_proto.ApplicationAuthMode.PRIVATE:
        auth_mode = "private"
    elif message.auth_mode is isolate_proto.ApplicationAuthMode.SHARED:
        auth_mode = "shared"
    else:
        raise ValueError(f"Unknown auth mode: {message.auth_mode}")

    return AliasInfo(
        alias=message.alias,
        revision=message.revision,
        auth_mode=auth_mode,
        keep_alive=message.keep_alive,
        max_concurrency=message.max_concurrency,
        max_multiplexing=message.max_multiplexing,
        active_runners=message.active_runners,
        min_concurrency=message.min_concurrency,
        concurrency_buffer=message.concurrency_buffer,
        concurrency_buffer_perc=message.concurrency_buffer_perc,
        machine_types=list(message.machine_types),
        request_timeout=message.request_timeout,
        startup_timeout=message.startup_timeout,
        valid_regions=list(message.valid_regions),
    )


@from_grpc.register(isolate_proto.RunnerInfo)
def _from_grpc_runner_info(message: isolate_proto.RunnerInfo) -> RunnerInfo:
    from isolate.server import definitions as worker_definitions

    external_metadata = worker_definitions.struct_to_dict(message.external_metadata)
    return RunnerInfo(
        runner_id=message.runner_id,
        in_flight_requests=message.in_flight_requests,
        expiration_countdown=(
            message.expiration_countdown
            if message.HasField("expiration_countdown")
            else None
        ),
        uptime=timedelta(seconds=message.uptime),
        external_metadata=external_metadata,
        revision=message.revision,
        alias=message.alias,
        state=RunnerState.from_proto(message.state),
    )


@from_grpc.register(isolate_proto.RegisterApplicationResult)
def _from_grpc_register_application_result(
    message: isolate_proto.RegisterApplicationResult,
) -> RegisterApplicationResult:
    return RegisterApplicationResult(
        logs=[from_grpc(log) for log in message.logs],
        result=(
            None
            if not message.HasField("result")
            else RegisterApplicationResultType(message.result.application_id)
        ),
        service_urls=from_grpc(message.service_urls)
        if message.HasField("service_urls")
        else None,
    )


@from_grpc.register(isolate_proto.HostedRunStatus)
def _from_grpc_hosted_run_status(
    message: isolate_proto.HostedRunStatus,
) -> HostedRunStatus:
    return HostedRunStatus(HostedRunState(message.state))


@from_grpc.register(isolate_proto.ServiceURLs)
def _from_grpc_service_urls(
    message: isolate_proto.ServiceURLs,
) -> ServiceURLs:
    return ServiceURLs(message.playground, message.run, message.queue, message.ws)


@from_grpc.register(isolate_proto.HostedRunResult)
def _from_grpc_hosted_run_result(
    message: isolate_proto.HostedRunResult,
) -> HostedRunResult[Any]:
    if message.return_value.definition:
        return_value = from_grpc(message.return_value)
    else:
        return_value = UNSET

    return HostedRunResult(
        message.run_id,
        from_grpc(message.status),
        logs=[from_grpc(log) for log in message.logs],
        result=return_value,
        service_urls=from_grpc(message.service_urls)
        if message.HasField("service_urls")
        else None,
    )


@dataclass
class MachineRequirements:
    machine_types: list[str]
    num_gpus: int | None = field(default=None)
    keep_alive: int = FAL_SERVERLESS_DEFAULT_KEEP_ALIVE
    base_image: str | None = None
    exposed_port: int | None = None
    scheduler: str | None = None
    scheduler_options: dict[str, Any] | None = None
    max_concurrency: int | None = None
    max_multiplexing: int | None = None
    min_concurrency: int | None = None
    concurrency_buffer: int | None = None
    concurrency_buffer_perc: int | None = None
    request_timeout: int | None = None
    startup_timeout: int | None = None

    def __post_init__(self):
        if isinstance(self.machine_types, str):
            self.machine_types = [self.machine_types]

        if not isinstance(self.machine_types, list):
            raise ValueError("machine_types must be a list of strings.")

        if not self.machine_types:
            raise ValueError("No machine type provided.")


@dataclass
class FalServerlessConnection:
    hostname: str
    credentials: Credentials

    _stack: ExitStack = field(default_factory=ExitStack)
    _stub: isolate_proto.IsolateControllerStub | None = None

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self._stack.close()

    def close(self):
        self._stack.close()

    @property
    def stub(self) -> isolate_proto.IsolateControllerStub:
        if self._stub:
            return self._stub

        options = self.credentials.server_credentials.base_options
        channel_creds = self.credentials.to_grpc()
        channel = self._stack.enter_context(
            grpc.secure_channel(
                target=self.hostname,
                credentials=channel_creds,
                options=list(options.items()),
            )
        )
        channel = grpc.intercept_channel(channel, TraceContextInterceptor())
        self._stub = isolate_proto.IsolateControllerStub(channel)
        return self._stub

    def create_user_key(self, scope: KeyScope, alias: str | None) -> tuple[str, str]:
        scope_proto = (
            isolate_proto.CreateUserKeyRequest.Scope.ADMIN
            if scope is KeyScope.ADMIN
            else isolate_proto.CreateUserKeyRequest.Scope.API
        )

        request = isolate_proto.CreateUserKeyRequest(scope=scope_proto, alias=alias)
        response = self.stub.CreateUserKey(request)
        return response.key_secret, response.key_id

    def list_user_keys(self) -> list[UserKeyInfo]:
        request = isolate_proto.ListUserKeysRequest()
        response: isolate_proto.ListUserKeysResponse = self.stub.ListUserKeys(request)
        return [
            UserKeyInfo(
                key.key_id,
                isolate_proto.datetime_from_timestamp(key.created_at),
                KeyScope.from_proto(key.scope),
                key.alias,
            )
            for key in response.user_keys
        ]

    def revoke_user_key(self, key_id) -> None:
        request = isolate_proto.RevokeUserKeyRequest(key_id=key_id)
        self.stub.RevokeUserKey(request)

    def define_environment(
        self, kind: str, **options: Any
    ) -> isolate_proto.EnvironmentDefinition:
        struct = isolate_proto.Struct()
        struct.update(options)

        return isolate_proto.EnvironmentDefinition(
            kind=kind,
            configuration=struct,
        )

    def register(
        self,
        function: Callable[..., ResultT],
        environments: list[isolate_proto.EnvironmentDefinition],
        application_name: str | None = None,
        auth_mode: Optional[AuthModeLiteral] = None,
        *,
        source_code: str | None = None,
        serialization_method: str = _DEFAULT_SERIALIZATION_METHOD,
        machine_requirements: MachineRequirements | None = None,
        metadata: dict[str, Any] | None = None,
        deployment_strategy: DeploymentStrategyLiteral,
        scale: bool = True,
        private_logs: bool = False,
        files: list[File] | None = None,
    ) -> Iterator[RegisterApplicationResult]:
        wrapped_function = to_serialized_object(function, serialization_method)
        if machine_requirements:
            wrapped_requirements = isolate_proto.MachineRequirements(
                # NOTE: backwards compatibility with old API
                machine_type=machine_requirements.machine_types[0],
                machine_types=machine_requirements.machine_types,
                num_gpus=machine_requirements.num_gpus,
                keep_alive=machine_requirements.keep_alive,
                base_image=machine_requirements.base_image,
                exposed_port=machine_requirements.exposed_port,
                scheduler=machine_requirements.scheduler,
                scheduler_options=to_struct(
                    machine_requirements.scheduler_options or {}
                ),
                max_concurrency=machine_requirements.max_concurrency,
                min_concurrency=machine_requirements.min_concurrency,
                concurrency_buffer=machine_requirements.concurrency_buffer,
                concurrency_buffer_perc=machine_requirements.concurrency_buffer_perc,
                max_multiplexing=machine_requirements.max_multiplexing,
                request_timeout=machine_requirements.request_timeout,
                startup_timeout=machine_requirements.startup_timeout,
            )
        else:
            wrapped_requirements = None

        if files:
            files = [
                isolate_proto.File(hash=file.hash, relative_path=file.relative_path)
                for file in files
            ]

        if auth_mode == "public":
            auth = isolate_proto.ApplicationAuthMode.PUBLIC
        elif auth_mode == "shared":
            auth = isolate_proto.ApplicationAuthMode.SHARED
        elif auth_mode == "private":
            auth = isolate_proto.ApplicationAuthMode.PRIVATE
        else:
            auth = None

        struct_metadata = None
        if metadata:
            struct_metadata = isolate_proto.Struct()
            struct_metadata.update(metadata)

        deployment_strategy_proto = DeploymentStrategy[
            deployment_strategy.upper()
        ].to_proto()

        request = isolate_proto.RegisterApplicationRequest(
            function=wrapped_function,
            environments=environments,
            machine_requirements=wrapped_requirements,
            application_name=application_name,
            auth_mode=auth,
            metadata=struct_metadata,
            deployment_strategy=deployment_strategy_proto,
            scale=scale,
            private_logs=private_logs,
            files=files,
            source_code=source_code,
        )
        for partial_result in self.stub.RegisterApplication(request):
            yield from_grpc(partial_result)

    def scale(self, application_name: str, max_concurrency: int | None = None) -> None:
        raise NotImplementedError

    def update_application(
        self,
        application_name: str,
        keep_alive: int | None = None,
        max_multiplexing: int | None = None,
        max_concurrency: int | None = None,
        min_concurrency: int | None = None,
        concurrency_buffer: int | None = None,
        concurrency_buffer_perc: int | None = None,
        request_timeout: int | None = None,
        startup_timeout: int | None = None,
        valid_regions: list[str] | None = None,
        machine_types: list[str] | None = None,
    ) -> AliasInfo:
        request = isolate_proto.UpdateApplicationRequest(
            application_name=application_name,
            keep_alive=keep_alive,
            max_multiplexing=max_multiplexing,
            max_concurrency=max_concurrency,
            min_concurrency=min_concurrency,
            concurrency_buffer=concurrency_buffer,
            concurrency_buffer_perc=concurrency_buffer_perc,
            request_timeout=request_timeout,
            startup_timeout=startup_timeout,
            valid_regions=valid_regions,
            machine_types=machine_types,
        )
        res: isolate_proto.UpdateApplicationResult = self.stub.UpdateApplication(
            request
        )
        return from_grpc(res.alias_info)

    def list_applications(
        self, application_name: str | None = None
    ) -> list[ApplicationInfo]:
        request = isolate_proto.ListApplicationsRequest(
            application_name=application_name
        )
        res: isolate_proto.ListApplicationsResult = self.stub.ListApplications(request)
        return [from_grpc(app) for app in res.applications]

    def delete_application(
        self,
        application_id: str,
    ) -> None:
        request = isolate_proto.DeleteApplicationRequest(application_id=application_id)
        self.stub.DeleteApplication(request)

    def rollout_application(
        self,
        application_name: str,
        force: bool = False,
    ) -> None:
        request = isolate_proto.RolloutApplicationRequest(
            application_name=application_name,
            force=force,
        )
        self.stub.RolloutApplication(request)

    def run(
        self,
        function: Callable[..., ResultT],
        environments: list[isolate_proto.EnvironmentDefinition],
        *,
        serialization_method: str = _DEFAULT_SERIALIZATION_METHOD,
        machine_requirements: MachineRequirements | None = None,
        setup_function: Callable[[], InputT] | None = None,
        files: list[File] | None = None,
    ) -> Iterator[HostedRunResult[ResultT]]:
        wrapped_function = to_serialized_object(function, serialization_method)
        if machine_requirements:
            wrapped_requirements = isolate_proto.MachineRequirements(
                # NOTE: backwards compatibility with old API
                machine_type=machine_requirements.machine_types[0],
                machine_types=machine_requirements.machine_types,
                num_gpus=machine_requirements.num_gpus,
                keep_alive=machine_requirements.keep_alive,
                base_image=machine_requirements.base_image,
                exposed_port=machine_requirements.exposed_port,
                scheduler=machine_requirements.scheduler,
                scheduler_options=to_struct(
                    machine_requirements.scheduler_options or {}
                ),
                max_concurrency=machine_requirements.max_concurrency,
                max_multiplexing=machine_requirements.max_multiplexing,
                min_concurrency=machine_requirements.min_concurrency,
                concurrency_buffer=machine_requirements.concurrency_buffer,
                concurrency_buffer_perc=machine_requirements.concurrency_buffer_perc,
                request_timeout=machine_requirements.request_timeout,
                startup_timeout=machine_requirements.startup_timeout,
            )
        else:
            wrapped_requirements = None
        if files:
            files = [
                isolate_proto.File(hash=file.hash, relative_path=file.relative_path)
                for file in files
            ]
        request = isolate_proto.HostedRun(
            function=wrapped_function,
            environments=environments,
            machine_requirements=wrapped_requirements,
            files=files,
        )
        if setup_function:
            request.setup_func.MergeFrom(
                to_serialized_object(setup_function, serialization_method)
            )
        stream = self.stub.Run(request)
        for partial_result in stream:
            res = from_grpc(partial_result)
            res.stream = stream
            yield res

    def create_alias(
        self,
        alias: str,
        revision: str,
        auth_mode: Optional[AuthModeLiteral],
    ) -> AliasInfo:
        if auth_mode == "public":
            auth = isolate_proto.ApplicationAuthMode.PUBLIC
        elif auth_mode == "shared":
            auth = isolate_proto.ApplicationAuthMode.SHARED
        elif auth_mode == "private":
            auth = isolate_proto.ApplicationAuthMode.PRIVATE
        else:
            auth = None

        request = isolate_proto.SetAliasRequest(
            alias=alias,
            revision=revision,
            auth_mode=auth,
        )
        res = self.stub.SetAlias(request)
        return from_grpc(res.alias_info)

    def delete_alias(self, alias: str) -> str | None:
        request = isolate_proto.DeleteAliasRequest(alias=alias)
        try:
            res: isolate_proto.DeleteAliasResult = self.stub.DeleteAlias(request)
            return res.revision
        except grpc.RpcError as e:
            if e.code() == grpc.StatusCode.NOT_FOUND:
                return None
            raise

    def list_aliases(self) -> list[AliasInfo]:
        request = isolate_proto.ListAliasesRequest()
        response: isolate_proto.ListAliasesResult = self.stub.ListAliases(request)
        return [from_grpc(alias) for alias in response.aliases]

    def list_alias_runners(
        self,
        alias: str,
        *,
        list_pending: bool = True,
        start_time: datetime | None = None,
    ) -> list[RunnerInfo]:
        kwargs = {"alias": alias, "list_pending": list_pending}
        if start_time:
            kwargs["start_time"] = isolate_proto.timestamp_from_datetime(start_time)

        request = isolate_proto.ListAliasRunnersRequest(**kwargs)
        response = self.stub.ListAliasRunners(request)
        return [from_grpc(runner) for runner in response.runners]

    def set_secret(self, name: str, value: str) -> None:
        request = isolate_proto.SetSecretRequest(name=name, value=value)
        self.stub.SetSecret(request)

    def delete_secret(self, name: str) -> None:
        request = isolate_proto.SetSecretRequest(name=name, value=None)
        self.stub.SetSecret(request)

    def list_secrets(self) -> list[ServerlessSecret]:
        request = isolate_proto.ListSecretsRequest()
        response = self.stub.ListSecrets(request)
        return [
            ServerlessSecret(
                name=secret.name,
                created_at=isolate_proto.datetime_from_timestamp(secret.created_time),
            )
            for secret in response.secrets
        ]

    def stop_runner(self, runner_id: str) -> None:
        request = isolate_proto.StopRunnerRequest(runner_id=runner_id)
        self.stub.StopRunner(request)

    def kill_runner(self, runner_id: str) -> None:
        request = isolate_proto.KillRunnerRequest(runner_id=runner_id)
        self.stub.KillRunner(request)

    def list_runners(self, start_time: datetime | None = None) -> list[RunnerInfo]:
        kwargs = {
            "list_pending": True,
        }
        if start_time:
            kwargs["start_time"] = isolate_proto.timestamp_from_datetime(start_time)

        request = isolate_proto.ListRunnersRequest(**kwargs)
        response = self.stub.ListRunners(request)
        return [from_grpc(runner) for runner in response.runners]



import functools
import random
import time
import traceback
from typing import Any, Callable, Literal, Optional

BackoffType = Literal["exponential", "fixed"]


def retry(
    max_retries: int = 3,
    base_delay: float = 1.0,
    max_delay: float = 60.0,
    backoff_type: BackoffType = "exponential",
    jitter: bool = False,
    should_retry: Optional[Callable[[Exception], bool]] = None,
) -> Callable:
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            retries = 0
            while retries < max_retries:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if should_retry is not None and not should_retry(e):
                        raise

                    retries += 1
                    print(f"Retrying {retries} of {max_retries}...")
                    if retries == max_retries:
                        print(f"Max retries reached. Raising exception: {e}")
                        traceback.print_exc()

                        raise e

                    if backoff_type == "exponential":
                        delay = min(base_delay * (2 ** (retries - 1)), max_delay)
                    else:  # fixed
                        delay = min(base_delay, max_delay)

                    if jitter:
                        delay *= random.uniform(0.5, 1.5)

                    time.sleep(delay)

        return wrapper

    return decorator



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import List, Optional

from fal.api import FAL_SERVERLESS_DEFAULT_URL, FalServerlessHost
from fal.sdk import (
    AliasInfo,
    Credentials,
    RunnerInfo,
)

from . import apps as apps_api
from . import deploy as deploy_api
from . import runners as runners_api


class _AppsNamespace:
    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def list(self, *, filter: str | None = None) -> List[AliasInfo]:
        return apps_api.list_apps(self.client, filter=filter)

    def runners(
        self, app_name: str, *, since=None, state: List[str] | None = None
    ) -> List[RunnerInfo]:
        return apps_api.apps_runners(self.client, app_name, since=since, state=state)

    def scale(
        self,
        app_name: str,
        *,
        keep_alive: int | None = None,
        max_multiplexing: int | None = None,
        max_concurrency: int | None = None,
        min_concurrency: int | None = None,
        concurrency_buffer: int | None = None,
        concurrency_buffer_perc: int | None = None,
        request_timeout: int | None = None,
        startup_timeout: int | None = None,
        machine_types: List[str] | None = None,
        regions: List[str] | None = None,
    ) -> apps_api.AliasInfo:
        return apps_api.scale_app(
            self.client,
            app_name,
            keep_alive=keep_alive,
            max_multiplexing=max_multiplexing,
            max_concurrency=max_concurrency,
            min_concurrency=min_concurrency,
            concurrency_buffer=concurrency_buffer,
            concurrency_buffer_perc=concurrency_buffer_perc,
            request_timeout=request_timeout,
            startup_timeout=startup_timeout,
            machine_types=machine_types,
            regions=regions,
        )

    def rollout(self, app_name: str, *, force: bool = False) -> None:
        return apps_api.rollout_app(self.client, app_name, force=force)


class _RunnersNamespace:
    def __init__(self, client: SyncServerlessClient):
        self.client = client

    def list(self, *, since=None) -> List[RunnerInfo]:
        return runners_api.list_runners(self.client, since=since)

    def stop(self, runner_id: str) -> None:
        return runners_api.stop_runner(self.client, runner_id)

    def kill(self, runner_id: str) -> None:
        return runners_api.kill_runner(self.client, runner_id)


@dataclass
class SyncServerlessClient:
    host: Optional[str] = None
    api_key: Optional[str] = None
    profile: Optional[str] = None
    team: Optional[str] = None

    def __post_init__(self) -> None:
        self.apps = _AppsNamespace(self)
        self.runners = _RunnersNamespace(self)

    # Top-level verbs
    def deploy(self, *args, **kwargs):
        return deploy_api.deploy(self, *args, **kwargs)

    # Internal helpers
    @property
    def _grpc_host(self) -> str:
        return self.host or FAL_SERVERLESS_DEFAULT_URL

    @property
    def _credentials(self) -> Credentials:
        from fal.sdk import FalServerlessKeyCredentials, get_default_credentials

        if self.api_key:
            if self.team:
                raise ValueError(
                    "Using explicit team with key credentials is not allowed"
                )
            try:
                key_id, key_secret = self.api_key.split(":", 1)
            except ValueError:
                raise ValueError("api_key must be in 'KEY_ID:KEY_SECRET' format")
            return FalServerlessKeyCredentials(key_id, key_secret)

        if self.profile:
            prev = os.environ.get("FAL_PROFILE")
            os.environ["FAL_PROFILE"] = self.profile
            try:
                return get_default_credentials(team=self.team)
            finally:
                if prev is None:
                    os.environ.pop("FAL_PROFILE", None)
                else:
                    os.environ["FAL_PROFILE"] = prev

        return get_default_credentials(team=self.team)

    def _create_host(self, *, local_file_path: str = "") -> FalServerlessHost:
        return FalServerlessHost(
            self._grpc_host,
            local_file_path=local_file_path,
            credentials=self._credentials,
        )



from __future__ import annotations

import asyncio
import inspect
import os
import pickle
import queue
import threading
import time
import traceback
import warnings
from collections.abc import AsyncIterator, Callable, Coroutine
from concurrent.futures import Future
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union

from fal.distributed.utils import (
    KeepAliveTimer,
    distributed_deserialize,
    distributed_serialize,
    encode_text_event,
    launch_distributed_processes,
)

if TYPE_CHECKING:
    import torch
    import torch.multiprocessing as mp
    from zmq.sugar.socket import Socket


class DistributedWorker:
    """
    A base class for distributed workers.
    """

    queue: queue.Queue[bytes]
    loop: asyncio.AbstractEventLoop
    thread: threading.Thread

    def __init__(
        self,
        rank: int = 0,
        world_size: int = 1,
    ) -> None:
        self.rank = rank
        self.world_size = world_size
        self.queue = queue.Queue()

        try:
            import uvloop

            self.loop = uvloop.new_event_loop()
        except ImportError:
            self.loop = asyncio.new_event_loop()

        self._start_thread()

    def _start_thread(self) -> None:
        """
        Start the thread.
        """
        self.thread = threading.Thread(target=self._run_forever, daemon=True)
        self.thread.start()

    def _run_forever(self) -> None:
        """
        Run the event loop forever.
        """
        asyncio.set_event_loop(self.loop)
        self.loop.run_forever()

    # Public API

    @property
    def device(self) -> torch.device:
        """
        :return: The device for the current worker.
        """
        import torch

        return torch.device(f"cuda:{self.rank}")

    @property
    def running(self) -> bool:
        """
        :return: Whether the event loop is running.
        """
        return self.thread.is_alive()

    def initialize(self, **kwargs: Any) -> None:
        """
        Initialize the worker.
        """
        import torch

        torch.cuda.set_device(self.device)
        self.rank_print(f"Initializing worker on device {self.device}")

        setup_start = time.time()
        future = self.run_in_worker(self.setup, **kwargs)
        future.result()
        setup_duration = time.time() - setup_start
        self.rank_print(f"Setup took {setup_duration:.2f} seconds")

    def add_streaming_result(
        self,
        result: Any,
        image_format: str = "jpeg",
        as_text_event: bool = False,
    ) -> None:
        """
        Add a streaming result to the queue.
        :param result: The result to add to the queue.
        """
        if as_text_event:
            serialized = encode_text_event(
                result, is_final=False, image_format=image_format
            )
        else:
            serialized = distributed_serialize(
                result, is_final=False, image_format=image_format
            )

        self.queue.put_nowait(serialized)

    def add_streaming_error(self, error: Exception) -> None:
        """
        Add an error to the queue.
        :param error: The error to add to the queue.
        """
        self.queue.put_nowait(
            distributed_serialize({"error": str(error)}, is_final=False)
        )

    def rank_print(self, message: str, debug: bool = False) -> None:
        """
        Print a message with the rank of the current worker.
        :param message: The message to print.
        :param debug: Whether to print the message as a debug message.
        """
        prefix = "[debug] " if debug else ""
        print(f"{prefix}[rank {self.rank}] {message}")

    def submit(self, coro: Coroutine[Any, Any, Any]) -> Future[Any]:
        """
        Submit a coroutine to the event loop.
        :param coro: The coroutine to submit to the event loop.
        :return: A future that will resolve to the result of the coroutine.
        """
        if not self.running:
            raise RuntimeError("Event loop is not running.")
        return asyncio.run_coroutine_threadsafe(coro, self.loop)

    def shutdown(self, timeout: Optional[Union[int, float]] = None) -> None:
        """
        Shutdown the event loop.
        :param timeout: The timeout for the shutdown.
        """
        try:
            self.run_in_worker(self.teardown).result()
        except Exception as e:
            self.rank_print(f"Error during teardown: {e}\n{traceback.format_exc()}")

        self.loop.call_soon_threadsafe(self.loop.stop)
        self.thread.join(timeout=timeout)

    async def _run_sync_in_executor(
        self,
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Run a synchronous function in the executor."""
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, partial(func, *args, **kwargs))

    def run_in_worker(
        self,
        func: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> Future[Any]:
        """
        Run a function in the worker.
        """
        if inspect.iscoroutinefunction(func):
            coro = func(*args, **kwargs)
        else:
            # Using in place of asyncio.to_thread
            # since it's not available in Python 3.8
            coro = self._run_sync_in_executor(func, *args, **kwargs)

        return self.submit(coro)

    # Overrideables

    def setup(self, **kwargs: Any) -> None:
        """
        Override this method to set up the worker.
        This method is called once per worker.
        """
        return

    def teardown(self) -> None:
        """
        Override this method to tear down the worker.
        This method is called once per worker.
        """
        return

    def __call__(self, streaming: bool = False, **kwargs: Any) -> Any:
        """
        Override this method to run the worker.
        """
        return {}


class DistributedRunner:
    """
    A class to launch and manage distributed workers.
    """

    zmq_socket: Optional[Socket[Any]]
    context: Optional[mp.ProcessContext]
    keepalive_timer: Optional[KeepAliveTimer]

    def __init__(
        self,
        worker_cls: type[DistributedWorker] = DistributedWorker,
        world_size: int = 1,
        master_addr: str = "127.0.0.1",
        master_port: int = 29500,
        worker_addr: str = "127.0.0.1",
        worker_port: int = 54923,
        timeout: int = 86400,  # 24 hours
        keepalive_payload: dict[str, Any] = {},
        keepalive_interval: Optional[Union[int, float]] = None,
        cwd: Optional[Union[str, Path]] = None,
        set_device: Optional[bool] = None,  # deprecated
    ) -> None:
        self.worker_cls = worker_cls
        self.world_size = world_size
        self.master_addr = master_addr
        self.master_port = master_port
        self.worker_addr = worker_addr
        self.worker_port = worker_port
        self.timeout = timeout
        self.cwd = cwd
        self.zmq_socket = None
        self.context = None
        self.keepalive_payload = keepalive_payload
        self.keepalive_interval = keepalive_interval
        self.keepalive_timer = None

        if set_device is not None:
            warnings.warn("set_device is deprecated and will be removed in the future.")

    def is_alive(self) -> bool:
        """
        Check if the distributed worker processes are alive.
        :return: True if the distributed processes are alive, False otherwise.
        """
        if self.context is None:
            return False
        for process in self.context.processes:
            if not process.is_alive():
                return False
        return True

    def terminate(self, timeout: Union[int, float] = 10) -> None:
        """
        Terminates the distributed worker processes.
        This method should be called to clean up the worker processes.
        """
        if self.context is not None:
            for process in self.context.processes:
                if process.is_alive():
                    process.terminate()
                    process.join(timeout=timeout)

    def gather_errors(self) -> list[Exception]:
        """
        Gathers errors from the distributed worker processes.

        This method should be called to collect any errors that occurred
        during execution.

        :return: A list of exceptions raised by the worker processes.
        """
        errors = []

        if self.context is not None:
            for error_file in self.context.error_files:
                if os.path.exists(error_file):
                    with open(error_file, "rb") as f:
                        error = pickle.loads(f.read())
                        errors.append(error)

                    os.remove(error_file)

        return errors

    def ensure_alive(self) -> None:
        """
        Ensures that the distributed worker processes are alive.
        If the processes are not alive, it raises an error.
        """
        if not self.is_alive():
            raise RuntimeError(
                f"Distributed processes are not running. Errors: {self.gather_errors()}"
            )

    def get_zmq_socket(self) -> Socket[Any]:
        """
        Returns a ZeroMQ socket of the specified type.
        :param socket_type: The type of the ZeroMQ socket.
        :return: A ZeroMQ socket.
        """
        if self.zmq_socket is not None:
            return self.zmq_socket

        import zmq
        import zmq.asyncio

        context = zmq.asyncio.Context()
        socket = context.socket(zmq.ROUTER)
        socket.bind(f"tcp://{self.worker_addr}:{self.worker_port}")
        self.zmq_socket = socket
        return socket

    def close_zmq_socket(self) -> None:
        """
        Closes the ZeroMQ socket.
        """
        if self.zmq_socket is not None:
            try:
                self.zmq_socket.close()
            except Exception as e:
                print(
                    f"[debug] Error closing ZeroMQ socket: {e}\n"
                    f"{traceback.format_exc()}"
                )
            self.zmq_socket = None

    def run(self, **kwargs: Any) -> None:
        """
        The main function to run the distributed worker.

        This function is called by each worker process spawned by
        `torch.multiprocessing.spawn`. This method must be synchronous.

        :param kwargs: The arguments to pass to the worker.
        """
        import torch.distributed as dist
        import zmq

        # Set up communication
        rank = int(os.environ["RANK"])
        context = zmq.Context()
        socket = context.socket(zmq.DEALER)
        socket.setsockopt(zmq.IDENTITY, str(rank).encode("utf-8"))
        socket.connect(f"tcp://{self.worker_addr}:{self.worker_port}")

        # Create and setup the worker
        worker = self.worker_cls(rank, self.world_size)
        try:
            worker.initialize(**kwargs)
        except Exception as e:
            worker.rank_print(
                f"Error during initialization: {e}\n{traceback.format_exc()}"
            )
            socket.send(b"EXIT")
            socket.close()
            return

        # Wait until all workers are ready
        socket.send(b"READY")
        dist.barrier()

        # Define execution methods to invoke from workers
        def execute(payload: bytes) -> Any:
            """
            Execute the worker function with the given payload synchronously.
            :param payload: The payload to send to the worker.
            :return: The result from the worker.
            """
            payload_dict = distributed_deserialize(payload)
            assert isinstance(payload_dict, dict)
            payload_dict["streaming"] = False

            try:
                future = worker.run_in_worker(worker.__call__, **payload_dict)
                result = future.result()
            except Exception as e:
                error_output = {"error": str(e)}
                worker.rank_print(
                    f"Error in execution: {error_output}\n{traceback.format_exc()}"
                )
                result = error_output

            dist.barrier()
            if worker.rank != 0:
                return

            socket.send(distributed_serialize(result, is_final=True))

        def stream(payload: bytes, as_text_events: bool) -> None:
            """
            Stream the result from the worker function with the given payload.
            :param payload: The payload to send to the worker.
            :return: An async iterator that yields the result from the worker.
            """
            payload_dict = distributed_deserialize(payload)
            assert isinstance(payload_dict, dict)
            payload_dict["streaming"] = True
            image_format = payload_dict.get("image_format", "jpeg")
            encoded_response: Optional[bytes] = None

            try:
                future = worker.run_in_worker(worker.__call__, **payload_dict)
                while not future.done():
                    try:
                        intermediate = worker.queue.get(timeout=0.1)
                        if intermediate is not None and worker.rank == 0:
                            socket.send(intermediate)  # already serialized
                    except queue.Empty:
                        pass
                result = future.result()
            except Exception as e:
                error_output = {"error": str(e)}
                worker.rank_print(
                    f"Error in streaming: {error_output}\n{traceback.format_exc()}"
                )
                if worker.rank == 0:
                    if as_text_events:
                        encoded_response = encode_text_event(error_output)
                    else:
                        encoded_response = distributed_serialize(error_output)
            else:
                if worker.rank == 0:
                    if as_text_events:
                        encoded_response = encode_text_event(
                            result, is_final=True, image_format=image_format
                        )
                    else:
                        encoded_response = distributed_serialize(
                            result, is_final=True, image_format=image_format
                        )

            dist.barrier()
            if worker.rank != 0:
                return

            if encoded_response is not None:
                socket.send(encoded_response)
            socket.send(b"DONE")

        # Runtime code
        if rank == 0:
            worker.rank_print("Master worker is ready to receive tasks.")
            while True:
                serialized_data = socket.recv()
                streaming = serialized_data[0] == ord("1")
                as_text_events = serialized_data[1] == ord("1")
                serialized_data = serialized_data[2:]
                params = [serialized_data, streaming, as_text_events]
                dist.broadcast_object_list(params, src=0)

                if serialized_data == b"EXIT":
                    worker.rank_print("Received exit payload, exiting.")
                    break

                if streaming:
                    stream(serialized_data, as_text_events)
                else:
                    execute(serialized_data)
        else:
            worker.rank_print("Worker waiting for tasks.")
            while True:
                try:
                    params = [None, None, None]
                    dist.broadcast_object_list(params, src=0)
                    payload, streaming, as_text_events = params  # type: ignore[assignment]
                    if payload == b"EXIT":
                        worker.rank_print("Received exit payload, exiting.")
                        break

                    if streaming:
                        stream(payload, as_text_events)  # type: ignore[arg-type]
                    else:
                        execute(payload)  # type: ignore[arg-type]
                except Exception as e:
                    worker.rank_print(f"Error in worker: {e}\n{traceback.format_exc()}")

        # Teardown
        worker.rank_print("Worker is tearing down.")
        try:
            worker.shutdown()
            worker.rank_print("Worker torn down successfully.")
        except Exception as e:
            worker.rank_print(f"Error during teardown: {e}\n{traceback.format_exc()}")

        socket.send(b"EXIT")
        socket.close()

    async def start(self, timeout: int = 1800, **kwargs: Any) -> None:
        """
        Starts the distributed worker processes.
        :param timeout: The timeout for the distributed processes.
        """
        import zmq

        if self.is_alive():
            raise RuntimeError("Distributed processes are already running.")

        self.context = launch_distributed_processes(
            self.run,
            world_size=self.world_size,
            master_addr=self.master_addr,
            master_port=self.master_port,
            timeout=self.timeout,
            cwd=self.cwd,
            **kwargs,
        )

        try:
            ready_workers: set[int] = set()
            socket = self.get_zmq_socket()
            start_time = time.perf_counter()

            while len(ready_workers) < self.world_size:
                try:
                    ident, msg = await socket.recv_multipart(flags=zmq.NOBLOCK)  # type: ignore[misc]

                    if msg != b"READY":
                        worker_id = ident.decode("utf-8")
                        worker_msg = msg.decode("utf-8")
                        raise RuntimeError(
                            f"Unexpected message from worker {worker_id}: {worker_msg}"
                        )

                    print(f"[debug] Worker {ident.decode('utf-8')} is ready.")
                    ready_workers.add(ident)
                except zmq.Again:
                    total_wait_time = time.perf_counter() - start_time
                    if total_wait_time > timeout:
                        raise TimeoutError(
                            f"Timeout reached after {timeout} seconds while "
                            f"waiting for workers to be ready."
                        )
                    await asyncio.sleep(0.5)
                self.ensure_alive()
        except Exception as e:
            print(f"[debug] Error during startup: {e}\n{traceback.format_exc()}")
            self.terminate(timeout=timeout)
            raise RuntimeError("Failed to start distributed processes.") from e

        print("[debug] All workers are ready and running.")

        # Start the keepalive timer
        self.maybe_start_keepalive()

    def keepalive(self, timeout: Optional[Union[int, float]] = 60.0) -> None:
        """
        Sends the keepalive payload to the worker.
        """
        # Cancel the keepalive timer
        self.maybe_cancel_keepalive()
        loop_thread = None
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            loop_thread = threading.Thread(target=loop.run_forever, daemon=True)
            loop_thread.start()

        future = asyncio.run_coroutine_threadsafe(
            self.invoke(self.keepalive_payload), loop
        )

        try:
            future.result(timeout=timeout)
        except Exception as e:
            print(f"[debug] Error during keepalive: {e}\n{traceback.format_exc()}")
            raise RuntimeError("Failed to run keepalive.") from e
        finally:
            if loop_thread is not None:
                loop.call_soon_threadsafe(loop.stop)
                loop_thread.join(timeout=timeout)
            # Restart the keepalive timer
            self.maybe_start_keepalive()

    def maybe_start_keepalive(self) -> None:
        """
        Starts the keepalive timer if it is set.
        """
        if self.keepalive_timer is None and self.keepalive_interval is not None:
            self.keepalive_timer = KeepAliveTimer(
                self.keepalive, self.keepalive_interval, start=True
            )

    def maybe_reset_keepalive(self) -> None:
        """
        Resets the keepalive timer if it is set.
        """
        if self.keepalive_timer is not None:
            self.keepalive_timer.reset()

    def maybe_cancel_keepalive(self) -> None:
        """
        Cancels the keepalive timer if it is set.
        """
        if self.keepalive_timer is not None:
            self.keepalive_timer.cancel()
            self.keepalive_timer = None

    async def stop(self, timeout: int = 10) -> None:
        """
        Stops the distributed worker processes.
        :param timeout: The timeout for the distributed processes to stop.
        """
        import zmq

        if not self.is_alive():
            raise RuntimeError("Distributed processes are not running.")

        self.maybe_cancel_keepalive()
        worker_exits: set[int] = set()
        socket = self.get_zmq_socket()
        await socket.send_multipart([b"0", b"00EXIT"])

        wait_start = time.perf_counter()
        while len(worker_exits) < self.world_size:
            try:
                ident, msg = await socket.recv_multipart(flags=zmq.NOBLOCK)  # type: ignore[misc]
                if msg == b"EXIT":
                    worker_exits.add(ident)
                    if len(worker_exits) == self.world_size:
                        print("[debug] All workers have exited.")
                        await asyncio.sleep(1)  # Allow time for cleanup
                        break

            except zmq.Again:
                # No messages available, continue waiting
                await asyncio.sleep(0.1)

            if time.perf_counter() - wait_start > timeout:
                print(
                    f"[debug] Timeout reached after {timeout} seconds, "
                    f"stopping waiting."
                )
                break

            if not self.is_alive():
                print("[debug] All workers have exited prematurely.")
                break

        if self.is_alive():
            print("[debug] Some workers did not exit cleanly, terminating them.")
            self.terminate(timeout=timeout)

        self.close_zmq_socket()

    async def stream(
        self,
        payload: dict[str, Any] = {},
        timeout: Optional[int] = None,
        streaming_timeout: Optional[int] = None,
        as_text_events: bool = False,
    ) -> AsyncIterator[Any]:
        """
        Streams the result from the distributed worker.
        :param payload: The payload to send to the worker.
        :param timeout: The timeout for the overall operation.
        :param streaming_timeout: The timeout in-between streamed results.
        :param as_text_events: Whether to yield results as text events.
        :return: An async iterator that yields the result from the worker.
        """
        import zmq

        self.ensure_alive()
        self.maybe_cancel_keepalive()  # Cancel until the streaming is done
        socket = self.get_zmq_socket()
        payload_serialized = distributed_serialize(payload, is_final=True)
        await socket.send_multipart(
            [b"0", b"1" + (b"1" if as_text_events else b"0") + payload_serialized]
        )

        start_time = time.perf_counter()
        last_yield_time = start_time
        yielded_once = False

        while True:
            iter_start_time = time.perf_counter()
            try:
                rank, response = await socket.recv_multipart(flags=zmq.NOBLOCK)  # type: ignore[misc]
            except zmq.Again:
                if timeout is not None and iter_start_time - start_time > timeout:
                    raise TimeoutError(f"Streaming timed out after {timeout} seconds.")
                if (
                    streaming_timeout is not None
                    and iter_start_time - last_yield_time > streaming_timeout
                ):
                    raise TimeoutError(
                        f"Streaming timed out after {streaming_timeout} "
                        f"seconds of inactivity."
                    )

                await asyncio.sleep(0.1)
                continue

            assert rank == b"0", "Expected response from worker with rank 0"

            if response == b"DONE":
                if not yielded_once:
                    raise RuntimeError("No data was yielded from the worker.")
                break

            if as_text_events:
                yield response
            else:
                yield distributed_deserialize(response)

            yielded_once = True
            last_yield_time = iter_start_time

        self.maybe_start_keepalive()  # Restart the keepalive timer

    async def invoke(
        self,
        payload: dict[str, Any] = {},
        timeout: Optional[int] = None,
    ) -> Any:
        """
        Invokes the distributed worker with the given payload.
        :param payload: The payload to send to the worker.
        :param timeout: The timeout for the overall operation.
        :return: The result from the worker.
        """
        import zmq

        self.ensure_alive()
        self.maybe_cancel_keepalive()  # Cancel until the invocation is done
        socket = self.get_zmq_socket()
        payload_serialized = distributed_serialize(payload, is_final=True)

        await socket.send_multipart([b"0", b"00" + payload_serialized])

        # Wait for the response from the worker
        start_time = time.perf_counter()
        while True:
            try:
                rank, response = await socket.recv_multipart(flags=zmq.NOBLOCK)  # type: ignore[misc]
                break  # Exit the loop if we received a response
            except zmq.Again:
                elapsed = time.perf_counter() - start_time
                if timeout is not None and elapsed > timeout:
                    raise TimeoutError(f"Invocation timed out after {timeout} seconds.")

                await asyncio.sleep(0.1)
                self.ensure_alive()

        self.maybe_start_keepalive()  # Restart the keepalive timer
        assert rank == b"0", "Expected response from worker with rank 0"
        return distributed_deserialize(response)

    async def __aenter__(self) -> DistributedRunner:
        """
        Enter the context manager.
        :return: The DistributedRunner instance.
        """
        await self.start()
        return self

    async def __aexit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        exc_traceback: Optional[traceback.StackSummary],
    ) -> None:
        """
        Exit the context manager.
        :param exc_type: The type of the exception raised, if any.
        :param exc_value: The value of the exception raised, if any.
        :param traceback: The traceback of the exception raised, if any.
        """
        try:
            await self.stop()
        except Exception as e:
            print(f"[debug] Error during cleanup: {e}\n{traceback.format_exc()}")



from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Sequence, Tuple, Union

import tomli


@lru_cache
def _load_toml(path: Union[Path, str]) -> Dict[str, Any]:
    with open(path, "rb") as f:
        return tomli.load(f)


@lru_cache
def _cached_resolve(path: Path) -> Path:
    return path.resolve()


@lru_cache
def find_project_root(srcs: Optional[Sequence[str]]) -> Tuple[Path, str]:
    """Return a directory containing .git, or pyproject.toml.

    That directory will be a common parent of all files and directories
    passed in `srcs`.

    If no directory in the tree contains a marker that would specify it's the
    project root, the root of the file system is returned.

    Returns a two-tuple with the first element as the project root path and
    the second element as a string describing the method by which the
    project root was discovered.
    """
    if not srcs:
        srcs = [str(_cached_resolve(Path.cwd()))]

    path_srcs = [_cached_resolve(Path(Path.cwd(), src)) for src in srcs]

    # A list of lists of parents for each 'src'. 'src' is included as a
    # "parent" of itself if it is a directory
    src_parents = [
        list(path.parents) + ([path] if path.is_dir() else []) for path in path_srcs
    ]

    common_base = max(
        set.intersection(*(set(parents) for parents in src_parents)),
        key=lambda path: path.parts,
    )

    for directory in (common_base, *common_base.parents):
        if (directory / ".git").exists():
            return directory, ".git directory"

        if (directory / "pyproject.toml").is_file():
            pyproject_toml = _load_toml(directory / "pyproject.toml")
            if "fal" in pyproject_toml.get("tool", {}):
                return directory, "pyproject.toml"

    return directory, "file system root"


def find_pyproject_toml(
    path_search_start: Optional[Tuple[str, ...]] = None,
) -> Optional[str]:
    """Find the absolute filepath to a pyproject.toml if it exists"""
    path_project_root, _ = find_project_root(path_search_start)
    path_pyproject_toml = path_project_root / "pyproject.toml"

    if path_pyproject_toml.is_file():
        return str(path_pyproject_toml)


def parse_pyproject_toml(path_config: str) -> Dict[str, Any]:
    """Parse a pyproject toml file, pulling out relevant parts for fal.

    If parsing fails, will raise a tomli.TOMLDecodeError.
    """
    pyproject_toml = _load_toml(path_config)
    config: Dict[str, Any] = pyproject_toml.get("tool", {}).get("fal", {})
    config = {k.replace("--", "").replace("-", "_"): v for k, v in config.items()}

    return config



import json
import os
import platform


def _doctor(args):
    import isolate

    import fal

    if args.output == "json":
        data = {
            "fal": fal.__version__,
            "isolate": isolate.__version__,
            "python": platform.python_version(),
            "platform": platform.platform(),
            "FAL_HOST": fal.flags.GRPC_HOST,
            "FAL_KEY": os.getenv("FAL_KEY", "").split(":")[0],
        }
        args.console.print(json.dumps(data))
    elif args.output == "pretty":
        from rich.table import Table

        table = Table(show_header=False, show_lines=False, box=None)
        table.add_column("name", no_wrap=True, style="bold")
        table.add_column("value", no_wrap=True)

        table.add_row("fal", fal.__version__)
        table.add_row("isolate", isolate.__version__)

        table.add_row("", "")
        table.add_row("python", platform.python_version())
        table.add_row("platform", platform.platform())

        table.add_row("", "")
        table.add_row("FAL_HOST", fal.flags.GRPC_HOST)
        table.add_row("FAL_KEY", os.getenv("FAL_KEY", "").split(":")[0])

        args.console.print(table)
    else:
        raise AssertionError(f"Invalid output format: {args.output}")


def add_parser(main_subparsers, parents):
    from .parser import get_output_parser

    doctor_help = "fal version and misc environment information."
    parser = main_subparsers.add_parser(
        "doctor",
        description=doctor_help,
        help=doctor_help,
        parents=[*parents, get_output_parser()],
    )
    parser.set_defaults(func=_doctor)
